﻿// (C) Copyright 2002-2011 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Collections.Specialized; // For StringCollection
using System.Threading;
using System.Diagnostics;
using System.ServiceProcess;

namespace AccoreConsoleDemo
{
    public partial class AccoreConsoleDemoForm : Form
    {
        private string drawingsFolderPath;     // Folder path of the drawings
        private string autocad2013InstallPath; // Acad 2013 install path. Used to locate accoreconsole.exe
        private StringCollection drawings;   // Collection of drawings found in the drawingsFolderPath
        private int maxWaitTime;             // Max duration in ms for which the accoreconsole.exe can be allowed to work in case of an error.
        private Image previewBitmapImage;    // Image generated for displaying drawing preview.
        private string AccoreConsoleAppPath; // Path of the accoreconsole.exe.
        private string selectedDrawingFile;  // Currently selected drawing that is selected in the list.

        public AccoreConsoleDemoForm()
        {
            InitializeComponent();

            // Initialization during startup
            Init();
        }

        /// <summary>
        /// Initialization of the accoreconsole.exe path variable
        /// </summary>
        private void Init()
        {
            // Get the path for the accoreconsole.exe
            autocad2013InstallPath = Properties.Settings.Default.AutoCAD2013InstallPath;
            AccoreConsoleAppPath = Path.Combine(Properties.Settings.Default.AutoCAD2013InstallPath, "accoreconsole.exe");

            drawings = new StringCollection();
            previewBitmapImage = null;

            maxWaitTime = Properties.Settings.Default.MaxWaitTime;
            drawingsFolderPath = Properties.Settings.Default.DrawingsFolderPath;
            this.StatusLabel.Text = "Drawings folder : " + drawingsFolderPath;
            
            #region Retrieve Accoreconsole.exe path
            // Read the registry to find the Acad 2013 install path.
            if (File.Exists(AccoreConsoleAppPath) == false)
            {
                autocad2013InstallPath = Utilities.GetAcad2013InstallPath();
                AccoreConsoleAppPath = Path.Combine(autocad2013InstallPath, "accoreconsole.exe");

                Properties.Settings.Default.AutoCAD2013InstallPath = autocad2013InstallPath;
                Properties.Settings.Default.Save(); 
                
                if (File.Exists(AccoreConsoleAppPath) == false)
                {
                    System.Windows.Forms.MessageBox.Show("AutoCAD 2013 path not found ! Please set the path using Settings.");
                }
            }
            #endregion

            #region Populate the Drawings list
            DrawingsList1.Items.Clear();
            DrawingsList2.Items.Clear();

            if (Directory.Exists(drawingsFolderPath) == false)
            {
                System.Windows.Forms.MessageBox.Show("Drawings folder path not found ! Please set the path using Settings.");
            }
            else
            {
                DirectoryInfo di = new DirectoryInfo(drawingsFolderPath);
                FileInfo[] fi = di.GetFiles("*.dwg");

                foreach (FileInfo info in fi)
                {
                    drawings.Add(info.FullName);
                    DrawingsList1.Items.Add(Path.GetFileName(info.FullName));
                    DrawingsList2.Items.Add(Path.GetFileName(info.FullName));
                }

                if (DrawingsList1.Items.Count > 0)
                    DrawingsList1.SelectedIndex = 0;

                if (DrawingsList2.Items.Count > 0)
                    DrawingsList2.SelectedIndex = 0;
            }
            #endregion
        }

        /// <summary>
        /// Generates preview image of the drawing file passed to it as parameter.
        /// The preview image is displayed in the PictureBox.
        /// </summary>
        /// <param name="drawingPath"></param>
        private void ShowPreview(string drawingPath)
        {
            // Clear the previous preview image.
            this.PreviewPictureBox.Image = null;
            if (previewBitmapImage != null)
            {
                previewBitmapImage.Dispose();
                previewBitmapImage = null;
            }

            // Delete the png file that is used for preview generation, if it already exists.
            string imagePath = Path.Combine(Utilities.AppPath, "DemoImage.png");
            if(Utilities.DeleteIfExists(imagePath) == false)
                return;

            string drawingFilePath = drawingPath;

            // Create the script file that we can pass to accoreconsole.exe
            // The script template is from the resource file.
            string scriptFilePath = Path.Combine(Utilities.AppPath, @"Scripts\Preview.scr");
            Utilities.CreateScriptFile(scriptFilePath, AccoreConsoleDemo.Resource1.PreviewScriptTemplate.Replace("<IMAGEPATH>", imagePath ));

            // Run Accoreconsole.exe
            Thread workerThread = Utilities.RunAccoreconsoleApp(AccoreConsoleAppPath, drawingFilePath, scriptFilePath, Properties.Settings.Default.MaxWaitTime);
            ShowProgress(ref workerThread);
            
            // Show the generated image as preview
            if (File.Exists(imagePath))
            {
                previewBitmapImage = Image.FromFile(imagePath, true);
                this.PreviewPictureBox.Image = previewBitmapImage;
            }
        }

        /// <summary>
        /// when a different drawing is selected in the listbox, update the preview image accordingly.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void DrawingsList_SelectedValueChanged(object sender, EventArgs e)
        {
            if(DrawingsList1.SelectedIndex != -1)
            {
                // Get the currently selected drawing and generate its preview
                selectedDrawingFile = Path.Combine(drawingsFolderPath, DrawingsList1.SelectedItem.ToString());
                ShowPreview(selectedDrawingFile);
            }
        }

        /// <summary>
        /// A script file to load a CRX file and invoke a custom command implemented in the CRX is first created.
        /// This script file is created using a template stored in the resource file.
        /// This script file is used with the accoreconsole.exe.
        /// </summary>
        /// <param name="drawingFilePath"></param>
        /// <param name="crxFilePath"></param>
        /// <param name="customCommandName"></param>
        private void RunCrxCustomCmd(string drawingFilePath, string crxFilePath, string customCommandName)
        {
            // Check if the CRX file exists.
            if (!File.Exists(crxFilePath))
            {
                System.Windows.Forms.MessageBox.Show("CRX file not found : " + crxFilePath);
                return;
            }

            // In the script template, replace the command name and the CRX file path before we can create the script file.
            crxFilePath = crxFilePath.Replace("\\", "\\\\");
            string script = AccoreConsoleDemo.Resource1.CrxCustomCommandScriptTemplate.Replace("<CRXPATH>", crxFilePath);
            script = script.Replace("<CUSTOMCOMMAND>", customCommandName);

            // Create the script file.
            // The script template is from the resource file.
            string scriptFilePath = Path.Combine(Utilities.AppPath, @"Scripts\CrxCustomCommand.scr");
            Utilities.CreateScriptFile(scriptFilePath, script);

            // Run Accoreconsole.exe
            Thread workerThread = Utilities.RunAccoreconsoleApp(AccoreConsoleAppPath, drawingFilePath, scriptFilePath, Properties.Settings.Default.MaxWaitTime);
            ShowProgress(ref workerThread);
        }

        /// <summary>
        /// A script file to load a .Net dll and invoke a custom command implemented in the .Net DLL is first created.
        /// This script file is created using a template stored in the resource file.
        /// This script file is used with the accoreconsole.exe.
        /// </summary>
        /// <param name="drawingFilePath"></param>
        /// <param name="dllFilePath"></param>
        /// <param name="customCommandName"></param>
        private void RunNetCustomCmd(string drawingFilePath, string dllFilePath, string customCommandName)
        {
            // Check if the .Net dll exists.
            if (!File.Exists(dllFilePath))
            {
                System.Windows.Forms.MessageBox.Show("Dll file not found : " + dllFilePath);
                return;
            }

            // In the script template, replace the command name and the .Net dll path before we can create the script file.
            string script = AccoreConsoleDemo.Resource1.DllCustomCommandScriptTemplate.Replace("<DLLPATH>",  dllFilePath);
            script = script.Replace("<CUSTOMCOMMAND>", customCommandName);

            // Create the script file.
            // The script template is from the resource file.
            string scriptFilePath = Path.Combine(Utilities.AppPath, @"Scripts\DllCustomCommand.scr");
            Utilities.CreateScriptFile(scriptFilePath, script);

            // Run Accoreconsole.exe
            Thread workerThread = Utilities.RunAccoreconsoleApp(AccoreConsoleAppPath, drawingFilePath, scriptFilePath, Properties.Settings.Default.MaxWaitTime);
            ShowProgress(ref workerThread);
        }

        /// <summary>
        /// This method invokes the accoreconsole.exe to execute a script file 
        /// that needs to be executed on a specified drawing file.
        /// The usage of accoreconsole.exe is :
        /// accoreconsole.exe /i <DrawingFilePath> /s <ScriptFilePath>
        /// </summary>
        /// <param name="drawingFilePath"></param>
        /// <param name="scrFilePath"></param>
        private void RunScript(string drawingFilePath, string scrFilePath)
        {
            // Check if the script file exists
            if (!File.Exists(scrFilePath))
            {
                System.Windows.Forms.MessageBox.Show("SCR file not found : " + scrFilePath);
                return;
            }

            // Run Accoreconsole.exe
            Thread workerThread = Utilities.RunAccoreconsoleApp(AccoreConsoleAppPath, drawingFilePath, scrFilePath, Properties.Settings.Default.MaxWaitTime);
            ShowProgress(ref workerThread);
        }

        /// <summary>
        /// Displays the settings form to take user inputs and saves the settings.
        /// When the settings dialog is dismissed, the drawings list is updated and the 
        /// preview is generated to reflect the changes in the settings.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SettingsButton_Click(object sender, EventArgs e)
        {
            SettingsForm sf = new SettingsForm();
            if (sf.ShowDialog() == DialogResult.OK)
            {
                bool updatePreview = false;
                if (drawingsFolderPath != Properties.Settings.Default.DrawingsFolderPath ||
                    AccoreConsoleAppPath != Path.Combine(Properties.Settings.Default.AutoCAD2013InstallPath, "accoreconsole.exe"))
                    updatePreview = true;

                AccoreConsoleAppPath = Path.Combine(Properties.Settings.Default.AutoCAD2013InstallPath, "accoreconsole.exe");
                drawingsFolderPath = Properties.Settings.Default.DrawingsFolderPath;
                maxWaitTime = Properties.Settings.Default.MaxWaitTime;

                if (File.Exists(AccoreConsoleAppPath) == false)
                {
                    System.Windows.Forms.MessageBox.Show("AutoCAD 2013 path not found ! Please set the path using Settings.");
                }

                if(updatePreview)
                {
                    #region Update the drawings list from the newly chosen drawings folder
                    DrawingsList1.Items.Clear();
                    DrawingsList2.Items.Clear();
                    drawings.Clear();

                    if (Directory.Exists(drawingsFolderPath) == false)
                    {
                        System.Windows.Forms.MessageBox.Show("Drawings folder path not found ! Please set the path using Settings.");
                    }
                    else
                    {
                        this.StatusLabel.Text = "Drawings folder : " + drawingsFolderPath;

                        DirectoryInfo di = new DirectoryInfo(drawingsFolderPath);
                        FileInfo[] fi = di.GetFiles("*.dwg");

                        foreach (FileInfo info in fi)
                        {
                            drawings.Add(info.FullName);
                            DrawingsList1.Items.Add(Path.GetFileName(info.FullName));
                            DrawingsList2.Items.Add(Path.GetFileName(info.FullName));
                        }

                        if (DrawingsList1.Items.Count > 0)
                            DrawingsList1.SelectedIndex = 0;

                        if (DrawingsList2.Items.Count > 0)
                            DrawingsList2.SelectedIndex = 0;
                    }
                    #endregion
                }
            }
        }

        /// <summary>
        /// Generates a script file for PDF export.
        /// Runs accoreconsole.exe on the selected drawing and using the PDF script file.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void PDFExportBtn_Click(object sender, EventArgs e)
        {
            // Erase the previous pdf file if it exists
            string pdfFilePath = Path.Combine(Utilities.AppPath, "Demo.pdf");
            if (Utilities.DeleteIfExists(pdfFilePath) == false)
                return;

            // Create the script file.
            // The script template is from the resource file.
            string scriptFilePath = Path.Combine(Utilities.AppPath, @"Scripts\PDF.scr");
            Utilities.CreateScriptFile(scriptFilePath, AccoreConsoleDemo.Resource1.PDFScriptTemplate.Replace("<REPLACEPATH>", pdfFilePath));

            // Run Accoreconsole.exe
            Thread workerThread = Utilities.RunAccoreconsoleApp(AccoreConsoleAppPath, selectedDrawingFile, scriptFilePath, Properties.Settings.Default.MaxWaitTime);
            ShowProgress(ref workerThread);

            // Open the generated pdf file using the default viewer (PDFReader ?)
            if (File.Exists(pdfFilePath))
            {
                Utilities.OpenDefaultViewer(pdfFilePath);
            }
        }

        /// <summary>
        /// Generates a script file for DWF export.
        /// Runs accoreconsole.exe on the selected drawing and using the DWF script file.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void DWFExportBtn_Click(object sender, EventArgs e)
        {
            // Erase the previous dwf file if it exists
            string dwfFilePath = Path.Combine(Utilities.AppPath, "Demo.dwf");
            if(Utilities.DeleteIfExists(dwfFilePath) == false)
                return;

            // Create the script file.
            // The script template is from the resource file.
            string scriptFilePath = Path.Combine(Utilities.AppPath, @"Scripts\DWF.scr");
            Utilities.CreateScriptFile(scriptFilePath, AccoreConsoleDemo.Resource1.DWFScriptTemplate.Replace("<REPLACEPATH>", dwfFilePath));

            // Run Accoreconsole.exe
            Thread workerThread = Utilities.RunAccoreconsoleApp(AccoreConsoleAppPath, selectedDrawingFile, scriptFilePath, Properties.Settings.Default.MaxWaitTime);
            ShowProgress(ref workerThread);

            // Open the generated DWF file using the default viewer (ADR ?)
            if (File.Exists(dwfFilePath))
            {
                Utilities.OpenDefaultViewer(dwfFilePath);
            }
        }

        /// <summary>
        /// On click of the "Run Command" button, the specified custom command implemented in a .CRX / .Net dll is
        /// invoked for all the selected drawings.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RunCmdBtn_Click(object sender, EventArgs e)
        {
            this.ConsoleOutputTextBox.Text = String.Empty;
            ListBox.SelectedObjectCollection selectedItems = this.DrawingsList2.SelectedItems;

            // Check for the selected drawings
            foreach (Object item in selectedItems)
            {
                string drawingFilePath = this.drawingsFolderPath + "\\" + item.ToString();

                // Check if the inputs are valid
                string crx_dllFilePath = CRX_DLLPathTextBox.Text;
                string customCommandName = CommandNameTextBox.Text;
                if (string.IsNullOrEmpty(drawingFilePath) ||
                    string.IsNullOrEmpty(crx_dllFilePath) ||
                    string.IsNullOrEmpty(customCommandName))
                {
                    return;
                }

                string extn = Path.GetExtension(crx_dllFilePath).ToLower();
                if (extn.Equals(".crx"))
                {// CRX file specified. Invoke the custom command from the crx.
                    RunCrxCustomCmd(drawingFilePath, crx_dllFilePath, customCommandName);
                    this.ConsoleOutputTextBox.Text = this.ConsoleOutputTextBox.Text + Utilities.GetLogText();
                }
                else if (extn.Equals(".dll"))
                {// .Net dll specified. Invoke the custom command from the .Net dll.
                    RunNetCustomCmd(drawingFilePath, crx_dllFilePath, customCommandName);
                    this.ConsoleOutputTextBox.Text = this.ConsoleOutputTextBox.Text + Utilities.GetLogText();
                }
            }
        }

        /// <summary>
        /// On click of the refresh button, generate the preview of the selected drawing.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RefreshButton_Click(object sender, EventArgs e)
        {
            ShowPreview(selectedDrawingFile);
        }

        /// <summary>
        /// Displays info about this demo app.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AboutDemoAppBtn_Click(object sender, EventArgs e)
        {
            String sContent = "AccoreConsole Demo for AutoCAD 2013\n\nDeveloper Technical Services\nAutodesk";
            System.Windows.Forms.MessageBox.Show(sContent, "About Accoreconsole Demo", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        /// <summary>
        /// On click of the "Run Script" button, the specified script file is executed for all the selected drawings.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RunScrBtn_Click(object sender, EventArgs e)
        {
            this.ConsoleOutputTextBox.Text = String.Empty;

            // Get the chosen drawings
            ListBox.SelectedObjectCollection selectedItems = this.DrawingsList2.SelectedItems;

            // Check for the selected drawings
            foreach (Object item in selectedItems)
            {
                string drawingFilePath = this.drawingsFolderPath + "\\" + item.ToString();

                // Check for valid inputs
                string scrFilePath = this.SCRPathTextBox.Text;
                if (string.IsNullOrEmpty(drawingFilePath) ||
                    string.IsNullOrEmpty(scrFilePath))
                {
                    return;
                }

                string extn = Path.GetExtension(scrFilePath).ToLower();
                if (extn.Equals(".scr"))
                {
                    // Execute the script on the drawing file.
                    RunScript(drawingFilePath, scrFilePath);

                    // append the output from the console window of accoreconsole.exe for display. 
                    this.ConsoleOutputTextBox.Text = this.ConsoleOutputTextBox.Text + Utilities.GetLogText();
                }
            }
        }

        /// <summary>
        /// Updates the progress bar until the workerthread passed as parameter is alive and kicking.
        /// Since we do not know how long it will take for accoreconsole.exe to finish, we display a marquee style progress bar.
        /// </summary>
        /// <param name="workerthread"></param>
        private void ShowProgress(ref Thread workerthread)
        {
            #region ProgressbarUpdate
            this.Cursor = Cursors.WaitCursor;
            this.ProgressBar.ProgressBar.Value = 0;
            this.ProgressBar.ProgressBar.Style = System.Windows.Forms.ProgressBarStyle.Marquee;
            do
            {
                System.Windows.Forms.Application.DoEvents();
            } while (workerthread.IsAlive);

            this.ProgressBar.ProgressBar.Style = System.Windows.Forms.ProgressBarStyle.Blocks;
            this.Cursor = Cursors.Arrow;
            #endregion
        }

        #region Browse buton click callbacks
        /// Displays a file selection dialog for selecting a CRX / .Net dll
        private void CRX_DllFileBrowseBtn_Click(object sender, EventArgs e)
        {
            OpenFileBrowserCRX_DLL.InitialDirectory = Utilities.AppPath;
            OpenFileBrowserCRX_DLL.Multiselect = false;
            OpenFileBrowserCRX_DLL.CheckFileExists = true;
            OpenFileBrowserCRX_DLL.DefaultExt = ".dll";
            OpenFileBrowserCRX_DLL.Filter = "CRX files (*.crx)|*.crx|.Net assembly files (*.dll)|*.dll";
            if (OpenFileBrowserCRX_DLL.ShowDialog() == DialogResult.OK)
            {
                CRX_DLLPathTextBox.Text = OpenFileBrowserCRX_DLL.FileName;
            }
        }

        /// Displays a file selection dialog for selecting a .SCR file
        private void SCRFileBrowseBtn_Click(object sender, EventArgs e)
        {
            OpenFileBrowserSCR.InitialDirectory = drawingsFolderPath;
            OpenFileBrowserSCR.Multiselect = false;
            OpenFileBrowserSCR.CheckFileExists = true;
            OpenFileBrowserSCR.DefaultExt = ".scr";
            OpenFileBrowserSCR.Filter = "Script files (*.scr)|*.scr";
            if (OpenFileBrowserSCR.ShowDialog() == DialogResult.OK)
            {
                this.SCRPathTextBox.Text = OpenFileBrowserSCR.FileName;
            }
        }
        #endregion

        private void ViewScriptBtn_Click(object sender, EventArgs e)
        {
            string scrFilePath = this.SCRPathTextBox.Text;
            if (string.IsNullOrEmpty(scrFilePath) == false)
            {
                Utilities.OpenDefaultViewer(scrFilePath); 
            }
        }
    }
}




