﻿Namespace Minesweeper

  'Interfaces that host app must implement to manage Minesweeper UI

#Region "Interface Definitions"

  'Provide input required by MinesweeperMgr

  Public Interface IMinesweeperInput
    'Requests host to get next game action
    ' e.g. Mark/Unmark/Uncover cell
    Function GetAction() As MinesweeperAction
  End Interface

  'Process output from MinesweeperMgr

  Public Interface IMinesweeperOutput
    'Tell host game is starting
    Sub StartGame()
    'Tell host when game is won or lost
    Sub EndGame(ByVal won As Boolean)
    'Tell host to update the value of a cell in the UI. Host returns
    ' false if it coouldn't update the cell.
    Function SetCell(ByVal cell As MineCell, ByVal row As Integer,
                     ByVal col As Integer) As Boolean
    'Tell host if there's an irrecoverable error
    Sub FatalError()
  End Interface

#End Region


  'Used by MinesweeperMgr internal error handling

  Public Class MinesweeperException
    Inherits Exception
  End Class


  'Used by MinesweeperAction to pass action type

  Public Enum GameAction
    Uncover
    Mark
    CancelGame
  End Enum

  'Used by host to provide input to MinesweeeprMgr -
  ' action type, and (row, col) of cell to act on.

  Public Structure MinesweeperAction

    Public Action As GameAction
    Public Row As Integer
    Public Col As Integer

    Public Sub New(ByVal act As GameAction, ByVal r As Integer,
                   ByVal c As Integer)

      Action = act
      Row = r
      Col = c

    End Sub

  End Structure

#Region "Enums and structures shared by MinesweeperMgr and host"

  'Used in MineCell structure to represent status of cell

  Public Enum CellStatus
    Covered
    Uncovered
    Marked
  End Enum

  'Represents a cell in the minefield (grid)

  Public Structure MineCell

    Public Status As CellStatus 'Covered/Uncovered/Marked
    Public Value As Integer 'No of neighboring cells containing bombs
    Public isBomb As Boolean 'Is this cell a bomb?

    Public Sub New(ByVal stat As CellStatus, ByVal val As Integer,
                    ByVal bomb As Boolean, ByVal marked As Boolean)

      Status = stat
      isBomb = bomb
      Value = val

    End Sub

  End Structure

#End Region


#Region "MinesweeperMgr class"

  Public Class MinesweeperMgr

    Private mRows As Integer 'Rows in grid
    Private mCols As Integer 'Columns in grid
    Private mMineArray(,) As MineCell 'The grid
    Private mNumMines As Integer 'Number of mines hidden in grid
    Private mNumCellsUncovered 'Number of cells currently uncovered
    Private mInputInterface As IMinesweeperInput
    Private mOutputInterface As IMinesweeperOutput


    'Instantiating MinesweeperMgr will initialize the game 
    ' and set it running. The class is meant to be used once and then
    ' discarded. Instantiate a new class to start a new game.
    'The host application must provide the input and output for this 
    ' class by implementing the IMinesweeperInput 
    ' and IMinesweeperOutput interfaces. Input and output have been 
    ' kept separate so  different classes can be used for each 
    ' if required.
    'MinesweeperMgr doen't know anything about how the host app 
    ' handles its input/output requests.

    Public Sub New(ByRef inputInterface As IMinesweeperInput,
                   ByRef outputInterface As IMinesweeperOutput,
                   ByVal rows As Integer, ByVal cols As Integer,
                   ByVal num As Integer)

      mInputInterface = inputInterface
      mOutputInterface = outputInterface
      InitMinefield(rows, cols, num)
      GameLoop()

    End Sub


    Private ReadOnly Property NumCellsUncovered() As Integer
      Get
        Return mNumCellsUncovered
      End Get
    End Property

    Private Function IncrementNumCellsUncovered() As Integer
      mNumCellsUncovered = mNumCellsUncovered + 1
      Return mNumCellsUncovered
    End Function

    Private Property MinefieldRows() As Integer
      Get
        Return mRows
      End Get
      Set(ByVal value As Integer)
        If value > 0 Then
          mRows = value
          If NumMines > mRows * MinefieldColumns Then
            NumMines = mRows * MinefieldColumns
          End If
        Else
          Throw New MinesweeperException
        End If
      End Set
    End Property

    Private Property MinefieldColumns() As Integer
      Get
        Return mCols
      End Get
      Set(ByVal value As Integer)
        If value > 0 Then
          mCols = value
          If NumMines > MinefieldRows * mCols Then
            NumMines = MinefieldRows * mCols
          End If
        Else
          Throw New MinesweeperException
        End If
      End Set
    End Property

    Private Property NumMines() As Integer
      Get
        Return mNumMines
      End Get
      Set(ByVal value As Integer)
        If mNumMines <= (mCols * mRows) Then
          mNumMines = value
        Else
          mNumMines = mCols * mRows
        End If
      End Set
    End Property

    'Private ReadOnly Property MineArray() As MineCell(,)
    '  Get
    '    Return mMineArray
    '  End Get
    'End Property


    'Uncover cell at specified location.
    'Returns true if uncovered cell  is a mine.

    Private Function UncoverCell(
      ByVal row As Integer, ByVal col As Integer) As Boolean

      'Only process covered cells- not marked ones
      '(Player would be annoyed if they accidentally cleared a 
      '  marked cell)
      If mMineArray(row, col).Status = CellStatus.Covered Then
        mMineArray(row, col).Status = CellStatus.Uncovered
        If Not mOutputInterface.SetCell(mMineArray(row, col),
                                        row, col) Then
          Throw New MinesweeperException
        End If
        IncrementNumCellsUncovered()
        ' If its a bomb, tell host the player lost
        If mMineArray(row, col).isBomb Then
          mOutputInterface.EndGame(False)
          Return True
          'If you cleared everything except the bombs, 
          ' tell host the player won
        ElseIf AllEmptyCellsUncovered() Then
          '
          mOutputInterface.EndGame(True)
          Return True
        End If
      End If
      Return False

    End Function

    'Toggle cell between Covered and Marked Status.
    'Ignores Uncovered cells.

    Private Sub MarkCell(
      ByVal row As Integer, ByVal col As Integer)

      If mMineArray(row, col).Status = CellStatus.Covered Then
        mMineArray(row, col).Status = CellStatus.Marked
      ElseIf mMineArray(row, col).Status = CellStatus.Marked Then
        mMineArray(row, col).Status = CellStatus.Covered
      End If
      If Not mOutputInterface.SetCell(mMineArray(row, col),
                                      row, col) Then
        Throw New MinesweeperException
      End If

    End Sub


    'Returns true if we've cleared all our non-mine cells

    Private Function AllEmptyCellsUncovered() As Boolean

      If NumCellsUncovered =
         MinefieldColumns * MinefieldRows - NumMines Then
        Return True
      Else
        Return False
      End If

    End Function


    Private Sub ResetNumCellsUncovered()

      mNumCellsUncovered = 0

    End Sub

    'Initialize grid, and put the mines in random locations

    Private Sub InitMinefield(
      ByVal rows As Integer, ByVal cols As Integer,
      ByVal num As Integer)

      Try


        If rows < 1 Or cols < 1 Or num < 1 Then
          Throw New MinesweeperException
        End If

        MinefieldRows = rows
        MinefieldColumns = cols
        If num > rows * cols Then
          NumMines = rows * cols
        Else
          NumMines = num
        End If

        'Initialize grid (array) to represent minefield

        ResetNumCellsUncovered()
        ReDim mMineArray(rows - 1, cols - 1)

        ' Add mines to grid (value of -1 means a mine is at
        ' that location)

        Randomize()
        Dim i As Integer = 0
        Do
          Dim rndRow As Integer = Rnd() * (MinefieldRows - 1)
          Dim rndCol As Integer = Rnd() * (MinefieldColumns - 1)
          If mMineArray(rndRow, rndCol).isBomb = False Then
            mMineArray(rndRow, rndCol).Value = -1
            mMineArray(rndRow, rndCol).isBomb = True
            mMineArray(rndRow, rndCol).Status = CellStatus.Covered
            i = i + 1
          End If
        Loop While i < num

        ' Now mines are added, we populate the rest of the grid
        ' with the numbers to indicate how many mines are in
        ' neighbouring(cells)

        For i = 0 To MinefieldRows - 1
          For j As Integer = 0 To MinefieldColumns - 1


            ' If this cell contains a mine then don't process it

            If Not mMineArray(i, j).isBomb = True Then

              Dim mineCounter As Integer = 0

              'Check grid cells around this one looking for mines ...
              ' i-1,j-1 | i,j-1 | i+1,j-1
              '   i-1,j |  i,j  | i+1,j
              ' i-1,j+1 | i,j+1 | i+1,j+1

              For k As Integer = -1 To 1
                For l As Integer = -1 To 1

                  ' Skip over cells outside bounds of minefield

                  If (i + k < 0) Or (i + k > MinefieldRows - 1) Or
                   (j + l < 0) Or (j + l > MinefieldColumns - 1) Then

                    Continue For
                  End If
                  'Don't include cell (i,j)
                  If k = 0 And l = 0 Then
                    Continue For
                  End If
                  If mMineArray(i + k, j + l).isBomb = True Then
                    mineCounter = mineCounter + 1
                  End If
                Next
              Next
              mMineArray(i, j).Value = mineCounter
              mMineArray(i, j).Status = CellStatus.Covered
            End If
            If Not mOutputInterface.SetCell(mMineArray(i, j),
                                            i, j) Then
              Throw New MinesweeperException
            End If
          Next
        Next

      Catch ex As MinesweeperException
        mOutputInterface.FatalError()
      End Try

    End Sub

    'This is the main game loop. It requests input from the host,
    ' acts on the input, updates its own data, and tells host to
    ' update its UI to reflect the change.

    Private Sub GameLoop()

      Dim bGameOver As Boolean

      'Tell host game is starting
      mOutputInterface.StartGame()

      Try

        'Loop until game ends or is cancelled. Possible input is
        ' Uncover a cell, toggles Mark on a cell, Cancel the game.

        Do
          Dim act As MinesweeperAction = mInputInterface.GetAction
          ValidateAction(act)
          Select Case act.Action
            Case GameAction.Uncover
              bGameOver = UncoverCell(act.Row, act.Col)
            Case GameAction.Mark
              MarkCell(act.Row, act.Col)
            Case GameAction.CancelGame
              bGameOver = True
          End Select

        Loop Until bGameOver

      Catch ex As MinesweeperException
        mOutputInterface.FatalError()
      End Try


    End Sub

    'Throws an exception if (act.Row, act.Col) is outside grid
    Private Sub ValidateAction(ByVal act As MinesweeperAction)

      If act.Row < 0 Or act.Row >= MinefieldRows Then
        Throw New MinesweeperException
      End If
      If act.Col < 0 Or act.Col >= MinefieldColumns Then
        Throw New MinesweeperException
      End If

    End Sub

  End Class

#End Region


End Namespace