// (C) Copyright 2011 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
//restricted rights notice below appear in all supporting 
//documentation.

//AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
//AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
//MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
//DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
//UNINTERRUPTED OR ERROR FREE.

//Use, duplication, or disclosure by the U.S. Government is subject to 
//restrictions set forth in FAR 52.227-19 (Commercial Computer
//Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
//(Rights in Technical Data and Computer Software), as applicable.

//Author: Philippe Leefsma, October 2011 - Autodesk Developer Technical Services

using System;
using System.IO;
using System.Reflection;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;

public delegate void StringHandler(string line);

public class AcadControl : UserControl 	
{
    private RichTextBox _TextBox;

    private string _prompt;
    private string _currentLine;
    private Timer _timer = new Timer();
    private StringHandler stringHandler;

    private static Interpreter interp;

    private int _historyIdx;
    private List<string> _history;

    static void ProcessLine(string line)
    {
        interp.ProcessLine(line);
    }    
    
    public AcadControl() 
    {
        InitializeComponent();

        _historyIdx = 0;
        _history = new List<string>();

        Interpreter.Console = new AcadConsole(this);
        interp = new Interpreter();

        interp.ReadIncludeFile("acad.csi");

        _prompt = "> ";
        stringHandler = new StringHandler(ProcessLine);

        _TextBox.WordWrap = false;
        
        Width = 750;
        Size = new Size(467, 400);

        _timer.Interval = 50;
        _timer.Tick += new EventHandler(Execute);
       
        _TextBox.KeyDown += new KeyEventHandler(TextBox_KeyDown);

        Write("\n" + _prompt + "Dynamic C# Interpreter for AutoCAD (1.0)");
        Write("\n" + _prompt);

        //filter = new CustomMessageFilter(this);
        //System.Windows.Forms.Application.AddMessageFilter(filter);
    }

    public void DisplayTerminalProperties()
    { 
        Form f = new Form();
        PropertyGrid grd =new PropertyGrid();
        grd.SelectedObject=_TextBox;
        grd.Dock=DockStyle.Fill;

        f.Controls.Add(grd);
        f.Text="Terminal Properties";
        f.Show();
    }

    void TextBox_KeyDown(object sender, KeyEventArgs e)
    {
        int lineNb = _TextBox.GetLineFromCharIndex(_TextBox.SelectionStart);

        if (lineNb != _TextBox.Lines.Length-1) 
        {
            e.SuppressKeyPress = true;
            e.Handled = true;
            return;
        }

        string line = _TextBox.Lines[lineNb];

        int idx1 = _TextBox.SelectionStart;
        int idx2 = _TextBox.GetFirstCharIndexFromLine(lineNb);

        if (idx1 - idx2 <= _prompt.Length)
        {
            _TextBox.SelectionStart = idx2 + _prompt.Length;

            if ((e.KeyCode == Keys.Back || e.KeyCode == Keys.Left))
            {
                e.SuppressKeyPress = true;
                e.Handled = true;
                return;
            }
        }

        if (e.KeyCode == Keys.Enter)
        {
            DelayedExecute(line);
            e.SuppressKeyPress = true;
            e.Handled = true;
            return;
        }

        if (e.KeyCode == Keys.Down && e.Shift)
        {
            DelayedExecute(line);
            e.SuppressKeyPress = true;
            e.Handled = true;
            return;
        }

        if (e.KeyCode == Keys.Up)
        {
            GetHistoryLine(true);
            e.SuppressKeyPress = true;
            e.Handled = true;
            return;
        }

        if (e.KeyCode == Keys.Down)
        {
            GetHistoryLine(false);
            e.SuppressKeyPress = true;
            e.Handled = true;
            return;
        }

        if (e.KeyCode == Keys.V && e.Control)
        {
            e.SuppressKeyPress = true;
            e.Handled = true;
            return;
        }
    }

    void DelayedExecute()
    {
        int lineNb = _TextBox.GetLineFromCharIndex(_TextBox.SelectionStart);

        if (lineNb != _TextBox.Lines.Length - 1)
        {
            return;
        }

        string line = _TextBox.Lines[lineNb];

        DelayedExecute(line); 
    }
    
    void DelayedExecute(string line) 
    {
        _currentLine = line;

        if( line == _prompt)
            return;
        
        _historyIdx = _history.Count;
        _history.Add(line);

        while (_currentLine.IndexOf(_prompt) == 0)
            _currentLine = _currentLine.Substring(_prompt.Length);

        _timer.Start();
    }

    void GetHistoryLine(bool restore_previous)
    {
        if(_history.Count == 0)
            return;

        int start = _TextBox.SelectionStart;
        int lineNb = _TextBox.GetLineFromCharIndex(start);

        string[] lines = _TextBox.Lines;

        if(restore_previous)
        {
            lines[lineNb] = _history[_historyIdx];

            _TextBox.Lines = lines;

            if (_historyIdx > 0)
                --_historyIdx;
        }
        else
        {
            if (_historyIdx < _history.Count - 1)
            {
                ++_historyIdx;
        
                lines[lineNb] = _history[_historyIdx];
                _TextBox.Lines = lines;
            }
            else
            {
                lines[lineNb] = _prompt;
                _TextBox.Lines = lines;
            }
        }

        _TextBox.SelectionStart = start;
    }
 
    void Execute(object sender,EventArgs e) 
    {
        _timer.Stop();
        stringHandler(_currentLine);
        _currentLine = string.Empty;
        Write("\n" + _prompt);
    }

    public void Write(string s) 
    {
        _TextBox.AppendText(s);
    }

    public void Clear()
    {
        //_historyIdx = 0;
        //_history.Clear();

        _TextBox.Clear();
    }

    private void InitializeComponent()
    {
        this._TextBox = new System.Windows.Forms.RichTextBox();
        this.SuspendLayout();
        // 
        // _TextBox
        // 
        this._TextBox.BackColor = System.Drawing.SystemColors.WindowText;
        this._TextBox.Dock = System.Windows.Forms.DockStyle.Fill;
        this._TextBox.Font = new System.Drawing.Font("Arial", 11.25F, System.Drawing.FontStyle.Bold, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
        this._TextBox.ForeColor = System.Drawing.Color.Lime;
        this._TextBox.Location = new System.Drawing.Point(0, 0);
        this._TextBox.Name = "_TextBox";
        this._TextBox.Size = new System.Drawing.Size(280, 300);
        this._TextBox.TabIndex = 0;
        this._TextBox.Text = "";
        // 
        // AcadControl
        // 
        this.Controls.Add(this._TextBox);
        this.Name = "AcadControl";
        this.Size = new System.Drawing.Size(280, 300);
        this.ResumeLayout(false);
    }

    //[System.Runtime.InteropServices.DllImport("user32.dll")]
    //public static extern IntPtr GetFocus();

    //CustomMessageFilter filter = null;

    //class CustomMessageFilter : System.Windows.Forms.IMessageFilter
    //{
    //    int WM_KEYDOWN = 0x0100;

    //    AcadControl _ctrl;

    //    public CustomMessageFilter(AcadControl ctrl)
    //    {
    //        _ctrl = ctrl;
    //    }

    //    public bool PreFilterMessage(ref System.Windows.Forms.Message m)
    //    {
    //        if (m.Msg == WM_KEYDOWN)
    //        {                
    //            if (m.WParam == (IntPtr)(13))  //Enter pressed
    //            {                    
    //                IntPtr hWnd = GetFocus();
    //                Control ctrl = Control.FromHandle(hWnd);

    //                if (ctrl == _ctrl._TextBox)
    //                {
    //                    _ctrl.DelayedExecute();
    //                }
    //            }
    //        }

    //        return false;
    //    }
    //}
}

class AcadConsole : IConsole
{
    AcadControl _ctrl;

    public AcadConsole(AcadControl ctrl)
    {
        _ctrl = ctrl;
        Enabled = false;
    }

    public string ReadLine()
    {
        return "";
    }

    public void Write(string s)
    {
        if(Enabled)
            _ctrl.Write(s);
    }

    public void Clear()
    {
        _ctrl.Clear();
    }

    public bool Enabled
    {
        get;
        set;
    }

    public void SetProperties()
    {
        _ctrl.DisplayTerminalProperties();
    }
}