﻿// (C) Copyright 2011 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.

using System;
using System.Linq;
using System.Collections.Generic;

using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.ApplicationServices;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.Geometry;
using Autodesk.AutoCAD.EditorInput;

[assembly: CommandClass(typeof(DynamicNetTest.MyCommands))]

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// DynamicNetTest is a sample project that illustrates the new dynamic 
// features of the AutoCAD 2013 .Net API
//
// Products: AutoCAD (and verticals) 2013 and higher
//
// Written by Philippe Leefsma, October 2011 - Developer Technical Services
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////
namespace DynamicNetTest
{
    public class MyCommands
    {
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Description: A comparison between classic and new dynamic syntax.
        //              The method renames layers with prefix "First Floor"
        //
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("ChgLayerNames")]
        public static void ChangeLayerNames()
        {
            Database db = HostApplicationServices.WorkingDatabase;

            using (Transaction tr = db.TransactionManager.StartTransaction())
            {
                LayerTable lt = (LayerTable)tr.GetObject(db.LayerTableId, OpenMode.ForRead);

                foreach (ObjectId ltrId in lt)
                {
                    LayerTableRecord ltr = (LayerTableRecord)tr.GetObject(ltrId, OpenMode.ForRead);

                    if (ltr.Name != "0")
                    {
                        ltr.UpgradeOpen();
                        ltr.Name = "First Floor " + ltr.Name;
                    }
                }

                tr.Commit();
            }
        }

        [CommandMethod("ChgLayerNamesDyna")]
        public static void ChangeLayerNamesDyna()
        {
            Database db = HostApplicationServices.WorkingDatabase;

            dynamic layers;

            layers = db.LayerTableId;

            foreach (dynamic l in layers)
            {
                if (l.Name != "0")
                    l.Name = "First Floor " + l.Name;
            }
        }

        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        //  Description: Dumps a few symbol tables
        //
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        static void DumpTable(dynamic symbolTable)
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;

            foreach (dynamic rec in symbolTable)
                doc.Editor.WriteMessage("\n - " + rec.Name);
        }
        
        [CommandMethod("DumpSymbolTable")]
        public static void DumpSymbolTable()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;

            doc.Editor.WriteMessage("\nBlocks: ");
            DumpTable(db.BlockTableId);

            doc.Editor.WriteMessage("\nLayers: ");
            DumpTable(db.LayerTableId);

            doc.Editor.WriteMessage("\nViews: ");
            DumpTable(db.ViewTableId);
        }

        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        //  Description: XRecord access and creation with dynamic syntax
        //
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        static public void xRecRead(string dicoName, string xRecName)
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            dynamic nod = db.NamedObjectsDictionaryId;

            try
            {
                var xr = nod[dicoName][xRecName];
                ed.WriteMessage("\n - " + xr.Data.ToString());
            }
            catch(Autodesk.AutoCAD.Runtime.Exception acadEx)
            { 
                if(acadEx.ErrorStatus == ErrorStatus.KeyNotFound)
                    ed.WriteMessage("\nDictionary or record does not exist");
            }
        }

        static public void xRecWrite(string dicoName, string xRecName)
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            dynamic nod = db.NamedObjectsDictionaryId;

            dynamic newdict = nod.SetAt(dicoName, new DBDictionary());
            dynamic record = newdict.SetAt(xRecName, new Xrecord());

            ResultBuffer rb = new ResultBuffer();
            rb.Add(new TypedValue((int)DxfCode.Text, "String1"));
            rb.Add(new TypedValue((int)DxfCode.Text, "String2"));

            record.Data = rb;
        }
        
        [CommandMethod("xRecReadCmd")]
        static public void xRecReadCmd()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Editor ed = doc.Editor;

            string dicoName;
            string xRecName;

            PromptResult pr1 = ed.GetString("\nEnter dictionary name: ");

            if (pr1.Status != PromptStatus.OK)
                return;

            dicoName = pr1.StringResult;


            PromptResult pr2 = ed.GetString("\nEnter xRecord name: ");

            if (pr2.Status != PromptStatus.OK)
                return;

            xRecName = pr2.StringResult;

            xRecRead(dicoName, xRecName);
        }

        [CommandMethod("xRecWriteCmd")]
        static public void xRecWriteCmd()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Editor ed = doc.Editor;

            string dicoName;
            string xRecName;

            PromptResult pr1 = ed.GetString("\nEnter dictionary name: ");

            if (pr1.Status != PromptStatus.OK)
                return;

            dicoName = pr1.StringResult;


            PromptResult pr2 = ed.GetString("\nEnter xRecord name: ");

            if (pr2.Status != PromptStatus.OK)
                return;

            xRecName = pr2.StringResult;

            xRecWrite(dicoName, xRecName);
        }

        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        //  Description: Adds an entity to working Database
        //
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        static dynamic AddToDatabase(Database db, Entity entity)
        {
            try
            {
                dynamic bt = db.BlockTableId;

                var modelspace = bt[BlockTableRecord.ModelSpace];

                return modelspace.AppendEntity(entity);
            }
            catch
            {
                return null;
            }
        }

        [CommandMethod("AddToDatabase")]
        public static void AddToDatabaseCmd()
        {
            dynamic dbEntity =
                AddToDatabase(
                    HostApplicationServices.WorkingDatabase,
                    new Circle(Point3d.Origin, Vector3d.ZAxis, 5.0));

            dbEntity.ColorIndex = 1;
        }
        
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        //  Description: A more comprehensive dynamic example with Linq queries
        //
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("DynamicDbTest")]
        public static void DynamicDbTest()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;

            //bootstrap dynamic typing. once db is 'dynamic' we use all intellisense and all expressions will be dynamic too
            dynamic db = HostApplicationServices.WorkingDatabase;

            //do some tests on layers
            var layers = db.LayerTableId;

            for (int i = 0; i < 100; ++i)
            {
                var newrec = layers.Add(new LayerTableRecord());
                newrec.Name = "Layer " + i.ToString();

                if (i % 4 == 0)
                    newrec.IsFrozen = true;
                if (i % 5 == 0)
                    newrec.IsOff = true;
            }

            //let's use some Linq
            //explicit cast to IEnumerable<dynamic> required
         
            //get the names of the layers that are frozen and off
            var frozenAndOffNames = from l in (IEnumerable<dynamic>)layers
                                    where l.IsFrozen == true && l.IsOff
                                    select l.Name;

            doc.Editor.WriteMessage("\nFrozen And Off Layers: ");

            foreach (string name in frozenAndOffNames)
            {
                doc.Editor.WriteMessage("\n - " + name);
            }

            //store layer5 so that we don't have to look it up each time through the loop
            var layer5 = layers["Layer 5"];
            
            //do some tests on blocks
            var modelspace = db.BlockTableId[BlockTableRecord.ModelSpace];

            for (int i = 0; i < 100; ++i)
            {
                dynamic newent;

                if (i % 3 == 0)
                    newent = modelspace.AppendEntity(new Line(new Point3d(i, 0, 0), new Point3d(0, i, 0)));
                else
                    newent = modelspace.AppendEntity(new Circle(new Point3d(i, i, 0), Vector3d.ZAxis, i));
                
                if (i % 5 == 0)
                    newent.LayerId = layer5;
            }

            //note the new IsKindOf operator. It is very efficient because it uses AcDbObjectId::objectClass
            //get the lines that are on layer5
            var lines = from e in (IEnumerable<dynamic>)modelspace
                        where e.IsKindOf(typeof(Line)) && e.Layer == "Layer5"
                        select e;

            //now some dictionary stuff
            var nod = db.NamedObjectsDictionaryId;
            var newdict = nod.SetAt("DynaTestDict", new DBDictionary());
            var record = newdict.SetAt("DynaTestRecord", new Xrecord());

            string data = "This is a dynamic test";

            record.Data = new ResultBuffer(new TypedValue((int)DxfCode.Text, data));

            //see if we can find the record we just added
            var records = from r in (IEnumerable<dynamic>)newdict
                          let v = r.Value
                          where v.IsKindOf(typeof(Xrecord)) && v.Data.AsArray()[0].Value == data
                          select v;

            doc.Editor.WriteMessage("\nXRecord Data: ");

            foreach (dynamic r in records)
                doc.Editor.WriteMessage("\n - " + r.Data.ToString());
            
            //a direct way:
            var xr = nod["DynaTestDict"]["DynaTestRecord"];
            doc.Editor.WriteMessage("\n - " + xr.Data.ToString());
        }
    }
}
